@inline function limit_slope(a, b, dx, prms)
  if prms.method === :none
    return 0.0
  elseif prms.method === :minmod
    return TCI.minmod(a, b)
  elseif prms.method === :muscl
    return TCI.minmod((a+b)/2, 2*a, 2*b)
  elseif prms.method === :superbee
    return TCI.minmod(TCI.maxmod(a,b),TCI.minmod(2*a,2*b))
  elseif prms.method === :van_albada
    c = dx^3
    return TCI.minmod( ((a^2+c^2)*b+(b^2+c^2)*a)/(a^2+b^2+c^2), 2*a, 2*b )
  elseif prms.method === :van_leer
    return TCI.minmod(2*a*b/(a+b), 2*a, 2*b)
  elseif prms.method === :tvb
    return TCI.minmod_M(dx, prms.tvb_M, (a+b)/2, 2*a, 2*b)
  elseif prms.method === :mc
    return dg1d.MC(b, a)
  else
    TODO(prms.method)
  end
end


#######################################################################
#                  slope limiters for FVElement mesh                  #
#######################################################################


# TODO Merge this with ScalarEq/rhs.jl:limit_slope!
function limit_slope!(δu, u, mesh::Mesh1d{<:FVElement}, @nospecialize(prms::NamedTuple))
  @unpack hrsc = prms
  if hrsc === :muscl
    limit_slope_muscl!(δu, u, mesh, prms)
  elseif hrsc === :fv_central
    # TODO Not sure if correct. The discussion in Hesthaven book 2018, sec 10.3 does not
    # specify how to compute the slope. The only requirement seems to be that
    # the slope obeys δu_j/h = du/dx(x_j) + O(h). And in the codes they provide they
    # seem to reuse the same slope limiters as for the MUSCL scheme etc.
    prms = (hrsc = :muscl, muscl_omega=0)
    limit_slope_muscl!(δu, u, mesh, prms)
  else
    TODO(hrsc)
  end
end


function limit_slope_muscl!(δu, u, mesh::Mesh1d{<:FVElement}, prms::NamedTuple)
  # assume that FV solution in jth cell [x_(j-1/2),x_(j+1/2)] can be approximated by
  #   u_j ≈ ̄u_j + δu_j (x-x_j)/h
  # where δu_j is the slope at x_j and
  #   x_j = x_j-1/2 + h/2
  #   h   = x_j+1/2 - x_j-1/2
  # Compare discussion in Hesthaven book 2018, sec 10.2
  @unpack muscl_omega = prms
  ω = muscl_omega
  # TODO Enabling upwinding in fv_muscl_burgers_sine (ω = 1.0) gives asymmetric results.
  # Do we need to compute two slopes like in the slope limiter implementations in EulerEq etc?
  K, = mesh.tree.dims
  @turbo for k = 2:K-1
    k₋, k₊ = k-1, k+1
    u₋, u₀, u₊ = u[k₋], u[k], u[k₊]
    Δ⁻u = u₀-u₋
    Δ⁺u = u₊-u₀
    # (10.12) in Hesthaven book 2018, sec 10.2
    δu[k] = (1+ω)/2*Δ⁻u + (1-ω)/2*Δ⁺u
  end
  if isperiodic(mesh)
    for k in (1,K)
      k₋, k₊ = periodic_index(k-1,K), periodic_index(k+1,K)
      u₋, u₀, u₊ = u[k₋], u[k], u[k₊]
      Δ⁻u = u₀-u₋
      Δ⁺u = u₊-u₀
      δu[k] = (1+ω)/2*Δ⁻u + (1-ω)/2*Δ⁺u
    end
  else
    TODO("bdry conditions")
  end
end


#######################################################################
#            slope limiter that preserves C^0 constraints             #
#######################################################################


function limit_slope_preserve_continuity!(var::AbstractArray, mask::AbstractArray, ws::Workspace, mesh::Mesh1d{<:DGElement})
  @toggled_assert mesh.element.kind === :modal_bspline1
  enter(ws) do

    @unpack x = get_static_variables(mesh)
    Npts = mesh.element.Npts
    K = mesh.tree.dims[1]
    tmp_var  = borrow(ws, Npts*K)
    avgs     = borrow(ws, Npts+1)
    δs       = borrow(ws, Npts+1)
    lim_avgs = borrow(ws, Npts-1)
    lim_δs   = borrow(ws, Npts-1)
    mat_A    = borrow(ws, (Npts-1)*(Npts-1))
    b        = borrow(ws, Npts-1)
    A        = reshape(mat_A,(Npts-1,Npts-1))

    for k in 1:K
      if k == 1 || k == K || mask[k] ≈ 0.0
        # skip bdry cells, but backup data because we restore the whole vector at once later
        idxs = cellindices(mesh, k)
        view(tmp_var, idxs) .= view(var, idxs)
        continue
      end

      idxs_lhs, idxs, idxs_rhs = cellindices.(Ref(mesh), (k-1,k,k+1))
      vvar_lhs, vvar, vvar_rhs = view.(Ref(var), (idxs_lhs,idxs,idxs_rhs))
      vx_lhs,   vx,   vx_rhs   = view.(Ref(x),   (idxs_lhs,idxs,idxs_rhs))
      vtmp_var                 = view(tmp_var, idxs)

      # determine averages and slopes, including neighbors
      v1, v2 = vvar_lhs[end-1], vvar_lhs[end]
      x1, x2 = vx_lhs[end-1], vx_lhs[end]
      avgs[1] = (v1+v2)/2; δs[1] = (v2-v1)/(x2-x1)
      v1, v2 = vvar_rhs[1], vvar_rhs[2]
      x1, x2 = vx_rhs[1], vx_rhs[2]
      avgs[end] = (v1+v2)/2; δs[end] = (v2-v1)/(x2-x1)
      for i in 1:Npts-1
        v1, v2 = vvar[i], vvar[i+1]
        x1, x2 = vx[i], vx[i+1]
        avgs[1+i] = (v1+v2)/2; δs[1+i] = (v2-v1)/(x2-x1)
      end

      # limit slopes
      for i in 2:Npts
        # lim_δs[i-1] = TCI.minmod(δs[i-1],δs[i],δs[i+1])
        x1, x2 = vx[i-1], vx[i]
        dx = x2-x1
        M = 0.1
        δl, δc, δr = δs[i-1], δs[i], δs[i+1]
        lim_δs[i-1] = TCI.minmod_M(dx, M, δc, δl, δr)
      end

      # recompute avgs so that continuity and total avg is preserved
      fill!(A, 0.0) # because we use lu! below
      for i in 1:Npts-2
        x1, x2, x3 = vx[i], vx[i+1], vx[i+2]
        dx1, dx2 = x2-x1, x3-x2
        A[i,i]   = 1.0
        A[i,i+1] = -1.0
        b[i] = - 0.5*dx1*lim_δs[i] - 0.5*dx2*lim_δs[i+1]
      end
      A[end,:] .= 1.0
      v_avgs = view(avgs, 2:Npts)
      b[end] = sum(v_avgs)

      # solve the equations
      F = lu!(A)
      ldiv!(lim_avgs, F, b)

      # resample data
      x1, x2 = vx[1], vx[2]
      dx = x2-x1
      vtmp_var[1] = lim_avgs[1] - 0.5*dx*lim_δs[1]
      x1, x2 = vx[end-1], vx[end]
      dx = x2-x1
      vtmp_var[end] = lim_avgs[end] + 0.5*dx*lim_δs[end]
      for i in 2:Npts-1
        x1, x2 = vx[i], vx[i+1]
        dx = x2-x1
        vtmp_var[i] = lim_avgs[i] - 0.5*dx*lim_δs[i]
      end
    end

    copyto!(var, tmp_var)
  end
end


function limit_slope_convex_hull!(var::AbstractArray, mask::AbstractArray,
                                   ws::Workspace, mesh::Mesh1d{<:DGElement})
  enter(ws) do

    @unpack x = get_static_variables(mesh)
    Npts = mesh.element.Npts
    K = mesh.tree.dims[1]
    tmp_var = borrow(ws, Npts*K)
    dvar = borrow(ws, Npts-1)
    w = 0.5

    for k in 1:K

      idxs = cellindices(mesh, k)
      vvar     = view(var, idxs)
      vx       = view(x, idxs)
      vtmp_var = view(tmp_var, idxs)

      if k == 1 || k == K || mask[k] ≈ 0.0
        # skip bdry cells, but backup data because we restore the whole vector at once later
        vtmp_var .= vvar
        continue
      end

      for i in 1:Npts-1
        dvar[i] = vvar[i+1]-vvar[i]
      end
      if abs(sum(sign, dvar)) ≈ Float64(Npts-1)
        vtmp_var .= vvar
        continue
      end

      idxs_lhs, idxs_rhs = cellindices.(Ref(mesh), (k-1,k+1))
      vvar_lhs, vvar_rhs = view.(Ref(var), (idxs_lhs,idxs_rhs))

      vtmp_var[1] = vvar[1] + (vvar_lhs[end]-vvar[1])*w
      vtmp_var[end] = vvar[end] + (vvar_rhs[1]-vvar[end])*w
      for i in 2:Npts-1
        ul, uc, ur = vvar[i-1], vvar[i], vvar[i+1]
        xl, xc, xr = vx[i-1], vx[i], vx[i+1]
        um = ul + (ur-ul)/(xr-xl)*(xc-xl)
        vtmp_var[i] = uc + (um-uc)*w
      end
    end

    copyto!(var, tmp_var)
  end
end
